@@ -38,9 +38,9 @@ char *alloca ();
 
 #if defined _LIBC
 # include <dirent.h>
-# define NAMLEN(dirent) _D_EXACT_NAMLEN(dirent)
+# define NAMLEN(dirent) _D_EXACT_NAMLEN (dirent)
 #else
-# if HAVE_DIRENT_H || defined _LIBC
+# if HAVE_DIRENT_H
 #  include <dirent.h>
 #  define NAMLEN(dirent) strlen ((dirent)->d_name)
 # else
@@ -60,6 +60,7 @@ char *alloca ();
 
 #include <errno.h>
 #include <ftw.h>
+#include <limits.h>
 #include <search.h>
 #include <stdlib.h>
 #include <string.h>
@@ -73,11 +74,6 @@ char *alloca ();
 # include <sys/stat.h>
 #endif
 
-#include <limits.h>
-#ifndef PATH_MAX
-# define PATH_MAX 1024
-#endif
-
 #if ! _LIBC && !HAVE_DECL_STPCPY && !defined stpcpy
 char *stpcpy ();
 #endif
@@ -97,20 +93,23 @@ char *stpcpy ();
 # define __closedir closedir
 # undef __fchdir
 # define __fchdir fchdir
+# undef __getcwd
+# define __getcwd(P, N) xgetcwd ()
+extern char *xgetcwd (void);
+# undef __mempcpy
+# define __mempcpy mempcpy
 # undef __opendir
 # define __opendir opendir
 # undef __readdir64
 # define __readdir64 readdir
+# undef __stpcpy
+# define __stpcpy stpcpy
 # undef __tdestroy
 # define __tdestroy tdestroy
 # undef __tfind
 # define __tfind tfind
 # undef __tsearch
 # define __tsearch tsearch
-# undef __stpcpy
-# define __stpcpy stpcpy
-# undef __mempcpy
-# define __mempcpy mempcpy
 # undef internal_function
 # define internal_function /* empty */
 # undef dirent64
@@ -119,12 +118,6 @@ char *stpcpy ();
 # define MAX(a, b) ((a) > (b) ? (a) : (b))
 #endif
 
-#ifndef _LIBC
-# undef __getcwd
-# define __getcwd(P, N) xgetcwd ()
-extern char *xgetcwd (void);
-#endif
-
 /* Arrange to make lstat calls go through the wrapper function
    on systems with an lstat function that does not dereference symlinks
    that are specified with a trailing slash.  */
@@ -155,6 +148,14 @@ int rpl_lstat (const char *, struct stat *);
 # define NFTW_FUNC_T __nftw_func_t
 #endif
 
+/* We define PATH_MAX if the system does not provide a definition.
+   This does not artificially limit any operation.  PATH_MAX is simply
+   used as a guesstimate for the expected maximal path length.
+   Buffers will be enlarged if necessary.  */
+#ifndef PATH_MAX
+# define PATH_MAX 1024
+#endif
+
 struct dir_data
 {
   DIR *stream;
@@ -426,7 +427,7 @@ process_entry (struct ftw_data *data, struct dir_data *dir, const char *name,
 
 	      if (result == 0 && (data->flags & FTW_CHDIR))
 		{
-		  /* Change back to parent directory.  */
+		  /* Change back to the parent directory.  */
 		  int done = 0;
 		  if (dir->stream != NULL)
 		    if (__fchdir (dirfd (dir->stream)) == 0)
@@ -656,6 +657,7 @@ ftw_startup (const char *dir, int is_nftw, void *func, int descriptors,
 						 * sizeof (struct dir_data *));
   memset (data.dirstreams, '\0', data.maxdir * sizeof (struct dir_data *));
 
+  /* PATH_MAX is always defined when we get here.  */
   dir_len = strlen (dir);
   data.dirbufsize = MAX (2 * dir_len, PATH_MAX);
   data.dirbuf = (char *) malloc (data.dirbufsize);
@@ -686,13 +688,13 @@ ftw_startup (const char *dir, int is_nftw, void *func, int descriptors,
   data.known_objects = NULL;
 
   /* Now go to the directory containing the initial file/directory.  */
-  if ((flags & FTW_CHDIR) && data.ftw.base > 0)
+  if (flags & FTW_CHDIR)
     {
       /* GNU extension ahead.  */
       cwd =  __getcwd (NULL, 0);
       if (cwd == NULL)
 	result = -1;
-      else
+      else if (data.ftw.base > 0)
 	{
 	  /* Change to the directory the file is in.  In data.dirbuf
 	     we have a writable copy of the file name.  Just NUL
@@ -768,10 +770,10 @@ ftw_startup (const char *dir, int is_nftw, void *func, int descriptors,
   /* Return to the start directory (if necessary).  */
   if (cwd != NULL)
     {
-      int saved_errno = errno;
+      save_err = errno;
       __chdir (cwd);
       free (cwd);
-      __set_errno (saved_errno);
+      __set_errno (save_err);
     }
 
   /* Free all memory.  */
